#ifndef MIXER_DISPLAY_H_INCLUDED
#define MIXER_DISPLAY_H_INCLUDED

void init_mixer_layout(void);
void display_card_info(void);
void display_view_mode(void);
void display_controls(void);
void compute_controls_layout(void);

#endif
